# Copyright 2022-2023 XProbe Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging
from typing import List, Optional

from ..._compat import Literal
from ..custom import ModelRegistry
from .core import ImageModelFamilyV2

logger = logging.getLogger(__name__)


class CustomImageModelFamilyV2(ImageModelFamilyV2):
    version: Literal[2] = 2
    model_id: Optional[str]  # type: ignore
    model_revision: Optional[str]  # type: ignore
    model_uri: Optional[str]
    controlnet: Optional[List["CustomImageModelFamilyV2"]]


UD_IMAGES: List[CustomImageModelFamilyV2] = []


class ImageModelRegistry(ModelRegistry):
    model_type = "image"

    def __init__(self):
        from .core import BUILTIN_IMAGE_MODELS

        super().__init__()
        self.models = UD_IMAGES
        self.builtin_models = list(BUILTIN_IMAGE_MODELS.keys())


def get_user_defined_images() -> List[ImageModelFamilyV2]:
    from ..custom import RegistryManager

    registry = RegistryManager.get_registry("image")
    return registry.get_custom_models()


def register_image(model_spec: CustomImageModelFamilyV2, persist: bool):
    from ..custom import RegistryManager

    registry = RegistryManager.get_registry("image")
    registry.register(model_spec, persist)


def unregister_image(model_name: str, raise_error: bool = True):
    from ..custom import RegistryManager

    registry = RegistryManager.get_registry("image")
    registry.unregister(model_name, raise_error)
